/** @file syn_sscon.cpp Standard synthesis consistency test */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2014  Matthias Leinfelder, Thomas Moor

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
   

#include "syn_sscon.h"
#include <stack>


namespace faudes {




/*
 ***************************************************************************************
 ***************************************************************************************

 The core algorithm is implemented as the function IsStdSynthesisConsistentUnchecked(...)
 The function assumes that the arguments satisfy their respecive conditions. For a
 detailed discussion on the algorithm see

 Moor, T.: Natural projections for the synthesis of 
 non-conflicting supervisory controllers, 
 Workshop on Discrete Event Systems (WODES), Paris, 2014.

 and

 Moor, T., Baier, Ch., Wittmann, Th.:
 Consistent abstractions for the purpose of supervisory control,
 Proc. 52nd IEEE Conference on Decision and Control (CDC), pp. 7291-7196, Firenze, 2013. 

 The initial version of this function was implemented by Matthias Leinfelder in course 
 of his bachelor thesis. 

 ***************************************************************************************
 ***************************************************************************************
 */


//IsStdSynthesisConsistentUnchecked (rPlantGen, rPlant0Gen, rCAlph)
bool IsStdSynthesisConsistentUnchecked(
  const Generator& rPlantGen, 
  const EventSet&  rCAlph, 
  const Generator&  rPlant0Gen) {  
    
  // prepare: extract relevant alpahbets
  const EventSet& sig  = rPlantGen.Alphabet();
  const EventSet& sigo = rPlant0Gen.Alphabet();
  const EventSet& sigc = rCAlph;
  EventSet sigco = sigc * sigo;
  EventSet siguc = sig - sigc;

  // prepare: construct rich plant representation by parallel composition with abstraction
  // (implementation of Paralle guarantees a trim result --- we rely on this)
  Generator gg0; 
  ProductCompositionMap gg0pmap;
  Parallel(rPlantGen,rPlant0Gen,gg0pmap,gg0);
    
  // prepare: construct a so-called trace generator with exactly the same reachable 
  // states as the rich plant gg0, but with only one path to reach each state; thus, 
  // each class of strings that reaches any particular state in gg0 has exaclty
  // one representative string in the generated language of the trace generator

  // initialize result and prepare the todo-stack for the loop invariant "todo states 
  // are reachable in trace by exactly one path"
  Generator trace;
  trace.InsEvents(sig);   
  std::stack<Idx> todo;      
  todo.push(gg0.InitState());       
  trace.InsInitState(gg0.InitState()); 
    
  // forward reachability search
  while(!todo.empty()){
    // pick state from todo
    Idx q = todo.top();
    todo.pop();
    // iterate all reachable states
    TransSet::Iterator tit,tit_end;               
    tit =  gg0.TransRelBegin(q);   
    tit_end =  gg0.TransRelEnd(q); 
    for(;tit != tit_end; ++tit){
      // by the invariant "tit->X2 not in trace" implies "tit->X2 is not on the stack"
      if(!trace.ExistsState(tit->X2)){ 
        todo.push(tit->X2);
        trace.InsState(tit->X2);
        trace.SetTransition(*tit);
      }
    }
  }
  // this completes the construction of the trace generator
    
  /*
  rPlantGen.GraphWrite("tmp_gen.jpg");
  rPlant0Gen.GraphWrite("tmp_gen0.jpg");
  gg0.GraphWrite("tmp_gg0.jpg");
  trace.GraphWrite("tmp_trace.jpg");
  */

  // prepare: backward reach via reverse transistion relation
  TTransSet<TransSort::X2EvX1> revtransrel; 
  gg0.TransRel().ReSort(revtransrel); 
    
  // initialize target states by marked states
  StateSet target = gg0.MarkedStates();
  // provide restricted target to focus backward reach (performance gain is debateble)
  StateSet rtarget = target;
     
  // grow target set by backward 'reach and test'
  while(true) {
    // prepare to sense when stuck / break on success
    Idx tsz=target.Size();
    if(tsz==gg0.Size()) break;

    // report
    FD_DF("IsStdSynthesisConsistent(..): targets for one-transition tests: #" << target.Size());

    // inner loop to grow by fast one-transition tests first
    while(true) {
      // allow for user interrupt
      FD_WPC(target.Size(),gg0.Size(),"IsStdSynthesisConsistent(): processing one-step tests"); 
      // prepare to sense when stuck / break on success
      Idx tszi=target.Size();
      if(tszi==gg0.Size()) break;
      // iterate over all relevant target states
      StateSet::Iterator sit = rtarget.Begin();
      StateSet::Iterator sit_end = rtarget.End();
      while(sit != sit_end){
        Idx x2=*sit;
        FD_DF("IsStdSynthesisConsistent(..): testing predecessors of target " << gg0.StateName(x2));
	// sense if all predecessors pass    
	bool allpass=true;
	// iterate over all target predecessors
	TTransSet<TransSort::X2EvX1>:: Iterator rit =revtransrel.BeginByX2(x2); 
	TTransSet<TransSort::X2EvX1>:: Iterator rit_end=revtransrel.EndByX2(x2); 
	for(;rit != rit_end; ++rit){
	  // predecessor technically passes if allready identified as target
	  if(target.Exists(rit->X1)) continue;
	  // test A and B from [ref2]: passes if uncontrollable/unobservable event leads to the target
	  if(siguc.Exists(rit->Ev)) {
	    target.Insert(rit->X1);
	    rtarget.Insert(rit->X1);
            FD_DF("IsStdSynthesisConsistent(..): one-transition test AB passed: " << gg0.StateName(rit->X1));
	    continue;
	  }
	  // test C adapted from [ref2]: predecessor passes if abstraction state is not marked and all successors lead to the target
          if(!rPlant0Gen.ExistsMarkedState(gg0pmap.Arg2State(rit->X1))){
            if(gg0.SuccessorStates(rit->X1) <= target) {
  	      target.Insert(rit->X1);
	      rtarget.Insert(rit->X1);
              FD_DF("IsStdSynthesisConsistent(..): one-transition test C passed: " << gg0.StateName(rit->X1));
	      continue;
	    } 
	  }
	  // record the fail
	  allpass=false;
	}
        // safely increment sit
        ++sit;
	// if all predecessors passed we can restrict the backward search
	if(allpass) {
          FD_DF("IsStdSynthesisConsistent(..): all predecessors of " << gg0.StateName(x2) << " have passed");
          rtarget.Erase(x2);
	}
      }
      // break inner loop when stuck
      if(target.Size()==tszi) break;
    }

    // break on success
    if(target.Size()==gg0.Size()) break;

    FD_DF("IsStdSynthesisConsistent(..): targets for multi-transition tests: #" << target.Size());
    // iterate over all relevant target states to apply the multi-transition test until success
    bool onepass=false;
    StateSet::Iterator sit =  rtarget.Begin();
    StateSet::Iterator sit_end =  rtarget.End();
    for(; (!onepass) && (sit != sit_end); ++sit){
      // iterate over all target predecessors
      TTransSet<TransSort::X2EvX1>:: Iterator rit =revtransrel.BeginByX2(*sit); 
      TTransSet<TransSort::X2EvX1>:: Iterator rit_end=revtransrel.EndByX2(*sit); 
      for(; (!onepass) && (rit != rit_end); ++rit){
	// cannot gain target if predecessor is already identified
	if(target.Exists(rit->X1)) continue;
        // allow for user interrupt
        FD_WPC(target.Size(),gg0.Size(),"IsStdSynthesisConsistent(): processing star-step tests"); 

        // prepare tests D/E from [ref2]/[ref1]

        // Ls: a) s
        Generator Ls(trace);                         
        Ls.SetMarkedState(rit->X1);                      
        // Ls: b) s . sig^*        --- all future of s
        Ls.ClrTransitions(rit->X1);
        SelfLoopMarkedStates(Ls,sig);
        // Ms: a) M                --- all successful future of s
        Generator Ms=gg0;
        Ms.InjectMarkedStates(target);
        // Ms: b) Ls . M
	LanguageIntersection(Ls,Ms,Ms);          
        // M0s: projection of Ms   --- all successful future of s from the perspective of the abstraction
        Generator M0s;
        Project(Ms,sigo,M0s);
        // Es0: a) Ms0 . sigo^*    --- evil specification to prevent succesful future of s
        Generator E0s=M0s;
        StateSet::Iterator ssit =  E0s.MarkedStatesBegin();
        StateSet::Iterator ssit_end =  E0s.MarkedStatesEnd();
        for(; ssit!=ssit_end; ++ssit) 
          E0s.ClrTransitions(*ssit);
        SelfLoopMarkedStates(E0s,sigo);
        // Es0: b) L0 - M0s . sigo^*       
        LanguageDifference(rPlant0Gen,E0s,E0s);
        Trim(E0s);

        // apply test D: does reachability of X1 contradict with the evil specification ?
        Generator Es=E0s;
        InvProject(Es,sig);
  	PrefixClosure(Es);
        if(EmptyLanguageIntersection(Ls,Es)) {
          // pass: evil specification renders X1 unreachable 
          FD_DF("IsStdSynthesisConsistent(..): multi-transition test D passed: " << gg0.StateName(rit->X1));
          onepass=true;
          target.Insert(rit->X1);
          rtarget.Insert(rit->X1);
          continue;
	}

        // apply test E: use the supremal evil supervisor to test whether reachability of X1 complies with E0s
        Generator K0s;
        SupRelativelyPrefixClosed(rPlant0Gen,E0s,E0s); 
        SupConNB(rPlant0Gen,sigco,E0s,K0s); 
	// test whether X1 is reachable under evil supervision
        Generator Ks=K0s;
        InvProject(Ks,sig);
  	PrefixClosure(Ks);
        Product(Ls,Ks,Ks);
        if(EmptyLanguageIntersection(Ls,Ks)) {
          // pass: evil supervisor renders X1 unreachable 
          FD_DF("IsStdSynthesisConsistent(..): multi-transition test E passed: " << gg0.StateName(rit->X1));
          onepass=true;
          target.Insert(rit->X1);
          rtarget.Insert(rit->X1);
	  continue;
   	}

      } // iterate target predecessors
    } // iterate targets

    // break outer loop if tests D/E did not gain one more target
    if(!onepass) break;

    // break outer loop if tests A/B are stuck 
    if(target.Size()==tsz) break;
    if(target.Size()==gg0.Size()) break;
  }

  // return success
  return target.Size()==gg0.Size();  
}
  
  


/*  
 ***************************************************************************************
 ***************************************************************************************

 Application Interface

 ***************************************************************************************
 ***************************************************************************************
 */


// IsStdSynthesisConsistent(rPlantGen, rCAlph, rPlant0Gen)
bool IsStdSynthesisConsistent(
  const Generator& rPlantGen, 
  const EventSet&  rCAlph, 
  const Generator&  rPlant0Gen) 
{  
  // test if the plant is deterministic
  if(!IsDeterministic(rPlantGen)){ 
    std::stringstream errstr;
    errstr << "Plant generator must be deterministic, " << "but is nondeterministic";
    throw Exception("IsStdSynthesisConsistent", errstr.str(), 501);
  }
  // test if the Alphabet with the controllable events matches the Alphabet of the Generator
  if(! (rCAlph <= rPlant0Gen.Alphabet()) ){
    std::stringstream errstr;
    errstr << "Controllable events must be a subset of the abstraction alphabet specified by \"" << rPlant0Gen.Name() << "\"";
    throw Exception("IsStdSynthesisConsistent", errstr.str(), 506);
  }  
  // test if the abstraction alphabet matches the plant alphabet
  const EventSet& sigo = rPlant0Gen.Alphabet();
  if(! (sigo <= rPlantGen.Alphabet() ) ){
    std::stringstream errstr;
    errstr << "Abstraction alphabet must be a subset of the plant alphabet pescified by \"" << rPlantGen.Name() << "\"";
    throw Exception("IsStdSynthesisConsistent", errstr.str(), 506);
  }  
    
  // test the consistency of the plant
  return IsStdSynthesisConsistentUnchecked(rPlantGen,rCAlph,rPlant0Gen);
}
  
  
// IsStdSynthesisConsistent(rPlantGen,rPlant0Gen)
bool IsStdSynthesisConsistent(	   
  const System& rPlantGen,
  const Generator& rPlant0Gen)
{
  // extract controllable events
  const EventSet& calph = rPlantGen.ControllableEvents();
  // pass on
  return IsStdSynthesisConsistent(rPlantGen,calph,rPlant0Gen);    
}


} // name space 
